@@ -142,6 +142,8 @@ class Repl {
         this.inputMode = target.querySelector(".snek-input-mode");
         this.statusBar = target.querySelector(".snek-status-bar");
         this.nav = target.querySelector(".snek-nav");
+        this.statusHistory = document.createElement("div");
+        this.statusHistory.className = "snek-status-history";
         this._setupEditor(this.inputBox);
 
         this.historySelection = 0;
@@ -474,6 +476,11 @@ class Repl {
             "popupActive"
         );
 
+        editor.addCommand(
+            KM.Alt | KC.US_SEMICOLON,
+            () => { this.editor.trigger(null, "editor.action.commentLine"); },
+        );
+
         this.editor = editor;
     }
 
@@ -576,8 +583,28 @@ class Repl {
         return elem;
     }
 
-    send(message) {
-        // Send a message back to SnekTalk
+    find_method(prefix, name, dflt) {
+        let method_name = `${prefix}_${name}`;
+        let method = this[method_name] || dflt;
+        return method && method.bind(this);
+    }
+
+    send(data) {
+        this.find_method("send", data.command, this.send_default)(data);
+    }
+
+    send_submit(data) {
+        let cmd = /^\/([^ ]+)( .*)?/.exec(data.expr);
+        let method = cmd && this.find_method("cmd", cmd[1], null);
+        if (method !== null) {
+            method(cmd[0], cmd[2]);
+        }
+        else {
+            this.send_default(data);
+        }
+    }
+
+    send_default(data) {
         if (this.closed) {
             this.setStatus({
                 type: "error",
@@ -585,10 +612,15 @@ class Repl {
             });
         }
         else {
-            this.socket.send(JSON.stringify(message));
+            this.socket.send(JSON.stringify(data));
         }
     }
 
+    cmd_status(fullexpr, arg) {
+        this.recv_echo({value: fullexpr});
+        this.append(this.statusHistory, "plain");
+    }
+
     append(elem, type) {
         let wrapper = document.createElement("div");
         wrapper.className = "snek-line snek-t-" + type;
@@ -596,7 +628,8 @@ class Repl {
         gutter.className = "snek-gutter snek-t-" + type;
         wrapper.appendChild(gutter);
         wrapper.appendChild(elem);
-        elem.className = "snek-result snek-t-" + type;
+        elem.classList.add("snek-result");
+        elem.classList.add("snek-t-" + type);
         this.pane.appendChild(wrapper);
         return wrapper;
     }
@@ -606,6 +639,9 @@ class Repl {
         this.statusBar.className =
             `snek-status-bar snek-status-${data.type} snek-status-flash-${data.type}`;
         this.statusBar.innerText = `${data.value} -- ${timestamp}`;
+        let statusCopy = document.createElement("div");
+        statusCopy.innerText = `${timestamp} ${data.value}`;
+        this.statusHistory.appendChild(statusCopy);
         setTimeout(
             () => {
                 this.statusBar.classList.remove(`snek-status-${data.type}`);
@@ -621,76 +657,83 @@ class Repl {
         );
     }
 
-    process_message(data) {
-        if (data.command == "resource") {
-            let elem = this.reify(data.value);
-            document.head.appendChild(elem);
+    recv_resource(data) {
+        let elem = this.reify(data.value);
+        document.head.appendChild(elem);
+    }
+
+    recv_result(data) {
+        let elem = this.reify(data.value);
+        let evalid = !data.evalid ? `E${++evalIdGen}` : data.evalid
+        let prbox = document.getElementById("pr-eval-" + evalid);
+        if (prbox === null) {
+            prbox = document.createElement("div");
+            prbox.id = "pr-eval-" + evalid;
+            this.append(prbox, "print");
         }
-        else if (data.command == "result") {
-            let elem = this.reify(data.value);
-            let evalid = !data.evalid ? `E${++evalIdGen}` : data.evalid
-            let prbox = document.getElementById("pr-eval-" + evalid);
-            if (prbox === null) {
-                prbox = document.createElement("div");
-                prbox.id = "pr-eval-" + evalid;
-                this.append(prbox, "print");
-            }
-            if (data.type === "statement") { }
-            else if (data.type === "print") {
-                prbox.appendChild(elem);
-            }
-            else {
-                this.append(elem, data.type);
-            }
-        }
-        else if (data.command == "echo") {
-            let ed = this.read_only_editor(data.value.trimEnd(), data.language || "python");
-            let elem = document.createElement("div");
-            elem.appendChild(ed);
-            this.append(elem, "echo");
-        }
-        else if (data.command == "response") {
-            let {resolve, reject} = this.$responseMap[data.response_id];
-            if (data.error) {
-                reject(data.error);
-            }
-            else {
-                resolve(data.value);
-            }
-        }
-        else if (data.command == "pastevar") {
-            let varname = data.value;
-            this.editor.trigger("keyboard", "type", {text: varname});
-            this.editor.focus();
-        }
-        else if (data.command == "status") {
-            this.setStatus(data);
-        }
-        else if (data.command == "set_nav") {
-            this.nav.innerHTML = "";
-            if (data.value) {
-                let elem = this.reify(data.value);
-                this.nav.appendChild(elem);
-            }
-        }
-        else if (data.command == "eval") {
-            eval(data.value);
-        }
-        else if (data.command == "set_mode") {
-            this.inputMode.innerHTML = data.html;
+        if (data.type === "statement") { }
+        else if (data.type === "print") {
+            prbox.appendChild(elem);
         }
         else {
-            console.error("Received an unknown command:", data.command);
+            this.append(elem, data.type);
         }
     }
 
+    recv_echo(data) {
+        let ed = this.read_only_editor(data.value.trimEnd(), data.language || "python");
+        let elem = document.createElement("div");
+        elem.appendChild(ed);
+        this.append(elem, "echo");
+    }
+
+    recv_response(data) {
+        let {resolve, reject} = this.$responseMap[data.response_id];
+        if (data.error) {
+            reject(data.error);
+        }
+        else {
+            resolve(data.value);
+        }
+    }
+
+    recv_pastevar(data) {
+        let varname = data.value;
+        this.editor.trigger("keyboard", "type", {text: varname});
+        this.editor.focus();
+    }
+
+    recv_status(data) {
+        this.setStatus(data);
+    }
+
+    recv_set_nav(data) {
+        this.nav.innerHTML = "";
+        if (data.value) {
+            let elem = this.reify(data.value);
+            this.nav.appendChild(elem);
+        }
+    }
+
+    recv_eval(data) {
+        eval(data.value);
+    }
+
+    recv_set_mode(data) {
+        this.inputMode.innerHTML = data.html;
+    }
+
+    recv_bad(data) {
+        console.error("Received an unknown command:", data.command);
+    }
+
     connect() {
         let port = window.location.port;
         let socket = new WebSocket(`ws://localhost:${port}/sktk?session=main`);
 
         socket.addEventListener('message', event => {
             let data = JSON.parse(event.data);
-            this.process_message(data);
+            this.find_method("recv", data.command, this.recv_bad)(data)
         });
 
         socket.addEventListener('error', event => {
