@@ -8,7 +8,8 @@ from jurigged import registry
 from jurigged.utils import glob_filter
 
 from . import runpy as snek_runpy
-from .evaluator import Evaluator
+from .evaluator import Evaluator, run_in_thread
+from .network import connect_to_existing
 from .server import serve
 
 
@@ -33,6 +34,26 @@ def cli():
     # [options: --no-watch]
     no_watch: Option & bool = default(False)
 
+    # Server port
+    # [alias: -p]
+    port: Option & int = default(None)
+
+    # Path to socket
+    # [alias: -S]
+    socket: Option & str = default(None)
+
+    # Hostname to connect to an existing instance
+    # [alias: -c]
+    connect: Option & str = default(False)
+
+    # Run the program in a thread
+    # [alias: -t]
+    thread: Option & bool = default(False)
+
+    if connect:
+        connect_to_existing(connect, port, socket)
+        return
+
     pattern = glob_filter(".")
     if no_watch:
         watch_args = None
@@ -42,12 +63,19 @@ def cli():
             "pattern": pattern,
         }
 
+    server_args = {
+        "port": port,
+        "sock": socket,
+    }
+
     sess = serve(
         watch_args=watch_args,
         template={"title": module or script or "snektalk"},
+        **server_args,
     )
     mod = None
     exc = None
+    run = None
 
     try:
         if module:
@@ -58,11 +86,15 @@ def cli():
             if ":" in module:
                 module, func = module.split(":", 1)
                 mod = importlib.import_module(module)
-                call = getattr(mod, func)
-                call()
+                run = getattr(mod, func)
 
             else:
-                mod = snek_runpy.run_module(module, run_name="__main__")
+                _, spec, code = snek_runpy._get_module_details(module)
+                if pattern(spec.origin):
+                    registry.prepare("__main__", spec.origin)
+                mod = ModuleType("__main__")
+                def run():
+                    snek_runpy.run_module(module, module_object=mod)
 
         elif script:
             path = os.path.abspath(script)
@@ -71,18 +103,26 @@ def cli():
                 # module resolution
                 registry.prepare("__main__", path)
             sys.argv[1:] = argv
-            mod = snek_runpy.run_path(path, run_name="__main__")
+            mod = ModuleType("__main__")
+            def run():
+                snek_runpy.run_path(path, module_object=mod)
 
         else:
             mod = ModuleType("__main__")
 
+        if run is not None:
+            if thread:
+                run_in_thread(run, session=sess)
+            else:
+                run()
+
     except Exception as exc:
         if sess is not None:
             sess.blt["$$exc_info"] = sys.exc_info()
-            sess.schedule(sess.send_result(exc, type="exception"))
+            sess.queue_result(exc, type="exception")
         else:
             raise
 
     finally:
-        if sess is not None and mod is not None:
-            Evaluator(mod, vars(mod), None, sess).loop()
+        if sess is not None:
+            Evaluator(mod, vars(mod) if mod else {}, None, sess).loop()
