@@ -31,7 +31,7 @@ class SnekTalkDb(bdb.Bdb):
         if frame.f_code.co_filename.startswith(_here):
             # Avoid running the debugger in snektalk code
             self.set_continue()
-        elif self.step_now:
+        elif getattr(self, "step_now", False):
             self.step_now = False
             self.set_step()
         else:
@@ -48,9 +48,9 @@ class SnekTalkDb(bdb.Bdb):
             )
         else:
             self.nav.js.update(
-                defn.live,
+                defn.codestring,
                 format_libpath(defn.filename),
-                defn.firstlineno,
+                defn.stashed.lineno,
                 frame.f_lineno,
             )
 
@@ -63,8 +63,11 @@ class SnekTalkDb(bdb.Bdb):
         while not self.proceed:
             self.show_frame(self.get_frame())
             with self.session.prompt(prompt=self.prompt, nav=self.nav) as cmd:
-                expr = cmd["expr"]
-                self.do(expr)
+                if cmd["command"] == "expr":
+                    expr = cmd["expr"]
+                    self.do(expr)
+                elif cmd["command"] == "noop":
+                    pass
 
     def error(self, message):
         print(H.i(message, style="color:red"))
