@@ -1,7 +1,9 @@
 import ast
 import functools
 import re
+import subprocess
 import sys
+import threading
 import time
 from types import ModuleType
 
@@ -10,6 +12,7 @@ from jurigged import CodeFile, registry
 from jurigged.recode import virtual_file
 
 from .fntools import find_fn
+from .session import current_session, new_evalid
 from .version import version
 
 cmd_rx = re.compile(r"/([^ ]+)( .*)?")
@@ -19,6 +22,17 @@ class StopEvaluator(Exception):
     pass
 
 
+def run_in_thread(fn, session=None):
+    def run():
+        with session.set_context():
+            with new_evalid():
+                fn()
+
+    session = session or current_session()
+    thread = threading.Thread(target=run, daemon=True)
+    thread.start()
+
+
 def safe_fail(fn):
     @functools.wraps(fn)
     def deco(self, *args, **kwargs):
@@ -27,14 +41,15 @@ def safe_fail(fn):
         except Exception as e:
             self.session.blt["_"] = e
             self.session.blt["$$exc_info"] = sys.exc_info()
-            self.session.schedule(self.session.send_result(e, type="exception"))
+            self.session.queue_result(e, type="exception")
 
     return deco
 
 
 def evaluate(expr, glb, lcl):
     filename = virtual_file("repl", expr)
-    cf = CodeFile(filename=filename, source=expr)
+    mname = glb.get("__name__", None)
+    cf = CodeFile(filename=filename, source=expr, module_name=mname)
     registry.cache[filename] = cf
 
     tree = ast.parse(expr)
@@ -59,7 +74,10 @@ def evaluate(expr, glb, lcl):
     else:
         rval = None
 
-    cf.discover(lcl, filename)
+    cf.code.set_globals(glb)
+    for v in lcl.values():
+        cf.associate(v)
+
     return rval
 
 
@@ -94,16 +112,14 @@ class Evaluator:
         assert isinstance(expr, str)
         expr = expr.lstrip()
 
-        self.session.schedule(
-            self.session.direct_send(command="echo", value=expr)
-        )
+        self.session.queue(command="echo", value=expr, process=False)
 
         result = self.eval(expr, glb, lcl)
         typ = "statement" if result is None else "expression"
 
         self.session.blt["_"] = result
 
-        self.session.schedule(self.session.send_result(result, type=typ))
+        self.session.queue_result(result, type=typ)
 
     @safe_fail
     def command_dir(self, expr, glb=None, lcl=None):
@@ -111,23 +127,21 @@ class Evaluator:
 
         expr = expr.lstrip()
 
-        self.session.schedule(
-            self.session.direct_send(command="echo", value=f"/dir {expr}")
-        )
+        self.session.queue(command="echo", value=f"/dir {expr}", process=False)
 
         result = hdir(self.eval(expr, glb, lcl))
         typ = "statement" if result is None else "expression"
 
         self.session.blt["_"] = result
 
-        self.session.schedule(self.session.send_result(result, type=typ))
+        self.session.queue_result(result, type=typ)
 
     @safe_fail
     def command_edit(self, expr, glb, lcl):
         expr = expr.lstrip()
         obj = self.eval(expr, glb, lcl)
-        self.session.schedule(
-            self.session.send_result(find_fn(obj), type="expression")
+        self.session.queue_result(
+            find_fn(obj, autofocus=True), type="expression"
         )
 
     @safe_fail
@@ -138,59 +152,71 @@ class Evaluator:
         glb = glb or self.glb
         lcl = lcl or self.lcl
         if expr:
-            self.session.schedule(
-                self.session.direct_send(command="echo", value=f"/debug {expr}")
+            self.session.queue(
+                command="echo", value=f"/debug {expr}", process=False
             )
             result = SnekTalkDb().runeval_step(expr, glb, lcl)
             typ = "statement" if result is None else "expression"
-            self.session.schedule(self.session.send_result(result, type=typ))
+            self.session.queue_result(result, type=typ)
 
         else:
             exc = self.session.blt.get("$$exc_info", None)
             if exc:
-                self.session.schedule(
-                    self.session.direct_send(
-                        command="echo",
-                        value=f"Debugging {exc[0].__qualname__}: {exc[1]}",
-                        language="text",
-                    )
+                self.session.queue(
+                    command="echo",
+                    value=f"Debugging {exc[0].__qualname__}: {exc[1]}",
+                    language="text",
+                    process=False,
                 )
                 tb = exc[2]
                 SnekTalkDb().interaction(tb.tb_frame, tb)
             else:
-                self.session.schedule(
-                    self.session.send_result(
-                        H.div("Last expression was not an exception"),
-                        type="exception",
-                    )
+                self.session.queue_result(
+                    H.div("Last expression was not an exception"),
+                    type="exception",
                 )
 
-    def command_quit(self, expr, glb, lcl):
-        self.session.schedule(
-            self.session.send_result(H.div("/quit"), type="echo",)
-        )
-        self.session.schedule(
-            self.session.send_result(
-                H.div("Quitting interpreter in ", self.format_modname()),
-                type="info",
+    @safe_fail
+    def command_shell(self, expr, glb, lcl):
+        expr = expr.lstrip()
+        self.session.queue(command="echo", value=f"//{expr}", process=False)
+        proc = subprocess.run(expr, shell=True, capture_output=True)
+        if proc.stdout:
+            self.session.queue_result(
+                H.div["snek-shellout"](proc.stdout.decode("utf8")),
+                type="expression",
             )
+        if proc.stderr:
+            self.session.queue_result(
+                H.div["snek-shellout"](proc.stderr.decode("utf8")),
+                type="exception",
+            )
+
+    @safe_fail
+    def command_thread(self, expr, glb, lcl):
+        run_in_thread(lambda: self.command_eval(expr, glb, lcl))
+
+    def command_quit(self, expr, glb, lcl):
+        self.session.queue_result(H.div("/quit"), type="echo")
+        self.session.queue_result(
+            H.div("Quitting interpreter in ", self.format_modname()),
+            type="info",
         )
         # Small delay so that the messages get flushed
         time.sleep(0.01)
         raise StopEvaluator()
 
     def missing(self, expr, cmd, arg, glb, lcl):
-        self.session.schedule(
-            self.session.direct_send(command="echo", value=expr)
-        )
-        self.session.schedule(
-            self.session.send_result(
-                H.div(f"Command '{cmd}' does not exist"), type="exception"
-            )
+        self.session.queue(command="echo", value=expr, process=False)
+        self.session.queue_result(
+            H.div(f"Command '{cmd}' does not exist"), type="exception"
         )
 
     def dispatch(self, expr, glb=None, lcl=None):
-        if match := cmd_rx.fullmatch(expr):
+        if expr.startswith("//"):
+            cmd = "shell"
+            arg = expr[2:]
+        elif match := cmd_rx.fullmatch(expr):
             cmd, arg = match.groups()
             if arg is None:
                 arg = ""
@@ -210,24 +236,25 @@ class Evaluator:
 
     def loop(self):
         pyv = sys.version_info
-        self.session.schedule(
-            self.session.send_result(
-                H.div(
-                    "Starting interpreter in ",
-                    self.format_modname(),
-                    H.br(),
-                    f"Snektalk {version} using Python {pyv.major}.{pyv.minor}.{pyv.micro}",
-                ),
-                type="info",
-            )
+        self.session.queue_result(
+            H.div(
+                "Starting interpreter in ",
+                self.format_modname(),
+                H.br(),
+                f"Snektalk {version} using Python {pyv.major}.{pyv.minor}.{pyv.micro}",
+            ),
+            type="info",
         )
         try:
             while True:
                 prompt = H.span["snek-input-mode-python"](">>>")
                 with self.session.prompt(prompt) as cmd:
-                    expr = cmd["expr"]
-                    if expr.strip():
-                        self.dispatch(expr)
+                    if cmd["command"] == "expr":
+                        expr = cmd["expr"]
+                        if expr.strip():
+                            self.dispatch(expr)
+                    elif cmd["command"] == "noop":
+                        pass
         except StopEvaluator:
             return
 
