@@ -16,6 +16,7 @@ from .registry import callback_registry
 _c = count(1)
 
 _current_session = ContextVar("current_session", default=None)
+_current_print_session = ContextVar("current_print_session", default=None)
 _current_evalid = ContextVar("current_evalid", default=None)
 
 
@@ -23,6 +24,10 @@ def current_session():
     return _current_session.get()
 
 
+def current_print_session():
+    return _current_print_session.get()
+
+
 @contextmanager
 def new_evalid():
     token = _current_evalid.set(next(_c))
@@ -33,7 +38,7 @@ def new_evalid():
 
 
 class Session:
-    def __init__(self, socket):
+    def __init__(self, socket=None):
         self.blt = vars(builtins)
         self.idmap = {}
         self.varcount = count(1)
@@ -41,27 +46,35 @@ class Session:
         self.sent_resources = set()
         self.last_prompt = ""
         self.last_nav = ""
-        self.command_queue = deque()
+        self.in_queue = deque()
+        self.out_queue = deque()
         self.semaphore = threading.Semaphore(value=0)
-        self.loop = asyncio.get_running_loop()
+        self._token = None
+        self._tokenp = None
 
     #############
     # Utilities #
     #############
 
-    def enter(self):
+    def enter(self, capture_print=True):
         self._token = _current_session.set(self)
+        if capture_print:
+            self._tokenp = _current_print_session.set(self)
 
     def exit(self):
         _current_session.reset(self._token)
+        if self._tokenp:
+            _current_print_session.reset(self._tokenp)
 
     @contextmanager
     def set_context(self):
         token = _current_session.set(self)
+        tokenp = _current_print_session.set(self)
         try:
             yield
         finally:
             _current_session.reset(token)
+            _current_print_session.reset(tokenp)
 
     def set_globals(self, glb):
         self.glb = glb
@@ -127,43 +140,48 @@ class Session:
 
     def next(self):
         self.semaphore.acquire()
-        return self.command_queue.popleft()
+        return self.in_queue.popleft()
 
     ###########
     # Sending #
     ###########
 
-    async def direct_send(self, **command):
-        """Send a command to the client."""
-        await self.socket.send(json.dumps(command))
+    def bind(self, socket):
+        self.loop = asyncio.get_running_loop()
+        self.socket = socket
+        self.sent_resources = set()
+        while self.out_queue:
+            self.schedule(self.send(**self.out_queue.popleft()))
+        self.submit({"command": "noop"})
 
-    async def send(self, **command):
+    async def send(self, process=True, **command):
         """Send a command to the client, plus any resources.
 
         Any field that is a Tag and contains resources will send
         resource commands to the client to load these resources.
         A resource is only sent once, the first time it is needed.
         """
-        resources = []
-        for k, v in command.items():
-            if isinstance(v, Tag):
-                resources.extend(v.collect_resources())
-                command[k] = str(v)
+        if process:
+            resources = []
+            for k, v in command.items():
+                if isinstance(v, Tag):
+                    resources.extend(v.collect_resources())
+                    command[k] = str(v)
 
-        for resource in resources:
-            if resource not in self.sent_resources:
-                await self.direct_send(command="resource", value=str(resource))
-                self.sent_resources.add(resource)
+            for resource in resources:
+                if resource not in self.sent_resources:
+                    await self.socket.send(
+                        json.dumps(
+                            {"command": "resource", "value": str(resource)}
+                        )
+                    )
+                    self.sent_resources.add(resource)
 
-        evalid = _current_evalid.get()
-        if evalid is not None:
-            command["evalid"] = evalid
+            evalid = _current_evalid.get()
+            if evalid is not None:
+                command["evalid"] = evalid
 
-        await self.direct_send(**command)
-
-    async def send_result(self, result, *, type):
-        type, html = self.represent(type, result)
-        await self.send(command="result", value=html, type=type)
+        await self.socket.send(json.dumps(command))
 
     def schedule(self, fn):
         self.loop.call_soon_threadsafe(lambda: self.loop.create_task(fn))
@@ -173,20 +191,30 @@ class Session:
 
         This queues the command using the session's asyncio loop.
         """
-        self.schedule(self.send(**command))
+        if self.socket is None or self.out_queue:
+            self.out_queue.append(command)
+        else:
+            self.schedule(self.send(**command))
+
+    def queue_result(self, result, *, type):
+        type, html = self.represent(type, result)
+        self.queue(command="result", value=html, type=type)
 
     ############
     # Commands #
     ############
 
+    def submit(self, data):
+        self.in_queue.append(data)
+        self.semaphore.release()
+
     async def recv(self, **command):
         cmd = command.pop("command", "none")
         meth = getattr(self, f"command_{cmd}", None)
         await meth(**command)
 
     async def command_submit(self, *, expr):
-        self.command_queue.append({"expr": expr})
-        self.semaphore.release()
+        self.submit({"command": "expr", "expr": expr})
 
     async def command_callback(self, *, id, response_id, arguments):
         try:
